# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   This script analyzes the experimental results from 
#   Andreoni & Samuelson (2006) using the methods in Bojinov et al (2020).
#   
#   The outcome is cooperation in the first period.

# Libraries and panel experiments functions -----------------------------------
library(tidyverse)
library(here)
library(latex2exp)

source(here("Code/panel_analysis_functions.R"))

# Load cleaned experimental results from Andreoni & Samuelson (2006) ----------
source(here("Code/andreoni-samuleson_clean-data_cooperation.R"))

timePeriods = 20 # number of periods in the experiment
unitIds = unique(experiment.data$unitId) # vector of unique unitIds
N = length(unitIds) # number of units

# Part 1: Compute lag-0, 0-step dynamic causal effects for two units -------- 
selected_units = c(10, 11) # user specified units

observedOutcomePlots = list()
htEstimatePlots = list()
for (unit in selected_units) {
  selected_unit_df = experiment.data %>% filter(unitId == unit)
  
  # Construct unit HT estimates
  unit_ht_wavg = unit.ht.wavg.est(expDF = experiment.data, unit = unit, p.lag = 0, normalize = FALSE)
  unit_ht_wavg_running = rep(0, length(unit_ht_wavg))
  for (t in 1:timePeriods) {
    unit_ht_wavg_running[t] = mean(unit_ht_wavg$unit.est[1:t])
  }
  selected_unit_df = selected_unit_df %>%
    mutate(ht_wavg = unit_ht_wavg$unit.est, 
           ht_wavg_running = unit_ht_wavg_running)
  
  # Plot HT estimates
  if (unit == selected_units[1]) {
    xtitle = "Game"
    ytitle = latex2exp::TeX("$\\tau$ estimate")
    title = "(a)"
  } else {
    xtitle = "Game"
    ytitle = ""
    title = "(b)"
  }
  
  htEstimatePlots[[as.character(unit)]] = ggplot(selected_unit_df, aes(x = period)) + 
    geom_line(aes(y = ht_wavg), size = 0.5, color = "black") +
    geom_line(aes(y = ht_wavg_running), size = 0.5, color = "black", linetype = 2) + 
    geom_hline(yintercept = unit_ht_wavg$unit.tavg.est, size = 0.5, color = "red", linetype = 2) + 
    geom_point(aes(y = ht_wavg, shape = factor(Wobs), color = factor(Wobs)), size = 2) + 
    scale_color_manual(values=c('#577ECB','#ED7D31')) + 
    labs(x = xtitle, y = ytitle, title = title, color = "W", shape = "W") + 
    theme_bw() + theme(legend.position = "bottom", plot.title = element_text(hjust = 0.5)) + ylim(-2.5, 2.5)
  
  rm(unit_ht_wavg, unit_ht_wavg_running, xtitle, ytitle)
}

# Plot results for two selected units
unitPlots = gridExtra::grid.arrange(grobs = list(htEstimatePlots[[1]], htEstimatePlots[[2]]), 
                                    widths = c(3, 3), nrow = 1)
ggsave(filename = here("Figures/AndreoniSamuelson-unit_plots-cooperation.jpeg"), plot = unitPlots, dpi = 150)

# Part 2: Compute cross-sectional lag-p, q-step averages for each period -------
# We compute this for p = 0, 1, 2, 3 and q = 0. 
p.lag.values = c(0, 1, 2, 3)

htCrossAvgPlots = list()
for (p.lag in p.lag.values) {
  # Compute cross-sectional average outcome and the cross-sectional average HT estimate for each period 
  period.ht.est = c(rep(NA, p.lag), rep(0, timePeriods - p.lag))
  period.ci.lb = c(rep(NA, p.lag), rep(0, timePeriods - p.lag))
  period.ci.ub = c(rep(NA, p.lag), rep(0, timePeriods - p.lag))
  
  for (t in (p.lag+1):timePeriods) {
    period.ht = time.normal.pval(expDF = experiment.data, time = t, p.lag = p.lag, returnCI = TRUE, alpha = 0.05)
    period.ht.est[t] = period.ht$cross.wavg.est
    period.ci.ub[t] = period.ht$CI[1]
    period.ci.lb[t] = period.ht$CI[2]
  }
  
  # Create dataframe of results
  cs_ht_df = tibble(period = 1:timePeriods,
                    period.ht.est = period.ht.est, 
                    ci.lb = period.ci.lb,
                    ci.ub = period.ci.ub
  )
  
  # Compute the total horvitz thompson estimate
  total.est = total.normal.pval(expDF = experiment.data, p.lag = p.lag, returnCI = TRUE, alpha = 0.05)
  
  # Plot Horvitz-Thompson estimates
  if (p.lag == 0) {
    xtitle = ""
    ytitle = latex2exp::TeX("$\\tau$ estimate")
    title = latex2exp::TeX("$p = 0")
  } else if (p.lag == 1) {
    xtitle = ""
    ytitle = ""
    title = latex2exp::TeX("$p = 1")
  } else if (p.lag == 2) {
    xtitle = "Game"
    ytitle = latex2exp::TeX("$\\tau$ estimate")
    title = latex2exp::TeX("$p = 2")
  } else {
    xtitle = "Game"
    ytitle = ""
    title = latex2exp::TeX("$p = 3")
  }
  
  htCrossAvgPlots[[as.character(p.lag)]] <- ggplot(cs_ht_df, aes(x = period)) + 
    geom_ribbon(aes(ymin = ci.lb, ymax = ci.ub), fill = "grey80", alpha = 0.5) + 
    geom_line(aes(y = period.ht.est), size = 0.5, color = "black", linetype = "dashed") + 
    geom_point(aes(y = period.ht.est), color = '#ED7D31', size = 2) +
    geom_hline(yintercept = total.est$total.wavg.est, color = "#577ECB", linetype = "solid", size = 0.5) + 
    geom_hline(yintercept = total.est$CI[1], color = "#577ECB", linetype = "dotted", size = 0.5) + 
    geom_hline(yintercept = total.est$CI[2], color = "#577ECB", linetype = "dotted", size = 0.5) + 
    labs(x = xtitle, y = ytitle, title = title) + 
    theme_bw() + theme(legend.position = "none", plot.title = element_text(hjust = 0.5)) + ylim(-1.5, 1.5)
  
  print(paste("p lag: ", p.lag))
  print(paste("total est: ", total.est$total.wavg.est))
  print(paste("p-value: ", total.est$p.val))
  
  rm(period.ht.est, period.ci.lb, period.ci.ub, cs_ht_df)
}
crossSectionPlots = gridExtra::grid.arrange(grobs = list(htCrossAvgPlots[[1]], htCrossAvgPlots[[2]], 
                                                         htCrossAvgPlots[[3]], htCrossAvgPlots[[4]]), 
                                            widths = c(3, 3), heights = c(2, 2), nrow = 2)
ggsave(filename = here("Figures/AndreoniSamuelson-crossSection_plots-cooperation.jpeg"), plot = crossSectionPlots, dpi = 150)

# Plot randomization distributions of the total dynamic causal effect estimates ---------
set.seed(1234567890)

# Need to construct the randomization distribution that respects the matches of units 
# i.e. each unit, partner pair needs to be assigned the same treatments. 
# As a result, we hand code a new randomization distribution function to do so.
N.sims = 10000
null.val = 0
randDistPlots = list()
for (p.lag in p.lag.values) {
  cat("Computing p.lag: ")
  cat(p.lag, "\n")
  
  rand.dist = rep(0, N.sims) 
  for (kk in 1:N.sims) {
    cat("Computing iteration: ")
    cat(kk)
    for (time in (p.lag+1):timePeriods) {
      expDFnew = experiment.data
      for (unit in unitIds) {
        # Draw new treatment assignment for unit
        unitDF = expDFnew[expDFnew$unitId == unit, ]
        Wnew = rbinom(p.lag+1, 1, prob = unitDF$pT[(time-p.lag):time])
        unitDF$Wobs[(time-p.lag):time] = Wnew
        expDFnew[expDFnew$unitId == unit, ] = unitDF
        
        # Assign the same treatment to partners 
        partnerDF = expDFnew[expDFnew$unitId == unitDF$partnerId[1], ]
        partnerDF$Wobs[(time-p.lag):time] = Wnew[((time-p.lag):time) - (time-p.lag-1)]
        expDFnew[expDFnew$unitId == unitDF$partnerId[1], ] = partnerDF
      }
      rand.dist[kk] = rand.dist[kk] + time.ht.wavg.est(expDF = expDFnew, time = time, p.lag = p.lag) + null.val
    }
    rand.dist[kk] = rand.dist[kk]/(timePeriods - p.lag)
    cat("\r")
  }
  htAvg_random = tibble(rand.dist = rand.dist)
  
  # Construct plot
  if (p.lag == 0) {
    xtitle = ""
    ytitle = "Density"
    title = latex2exp::TeX("$p = 0")
  } else if (p.lag == 1) {
    xtitle = ""
    ytitle = ""
    title = latex2exp::TeX("$p = 1")
  } else if (p.lag == 2) {
    xtitle = latex2exp::TeX("$\\tau$ estimate")
    ytitle = "Density"
    title = latex2exp::TeX("$p = 2")
  } else {
    xtitle = latex2exp::TeX("$\\tau$ estimate")
    ytitle = ""
    title = latex2exp::TeX("$p = 3")
  }
  
  randDistPlots[[as.character(p.lag)]] = ggplot(data.frame(rand.dist = htAvg_random$rand.dist), aes(x = rand.dist, y = ..density..)) + 
    geom_histogram(bins = 100, fill = "#577ECB", colour = "black", boundary = 0) + 
    geom_vline(xintercept = total.ht.wavg.est(expDF = experiment.data, p.lag = p.lag), color = '#ED7D31', size = 1, linetype = "dashed") + 
    labs(x = xtitle, y = ytitle, title = title) + theme_bw() + 
    theme(plot.title = element_text(hjust = 0.5))
  
  ggsave(filename = here(paste("Figures/AndreoniSamuelson-randDist_p", p.lag, "-cooperation.jpeg", sep = "")), 
         plot = randDistPlots[[as.character(p.lag)]])
  saveRDS(object = htAvg_random, file = here(paste("Temp/AndreoniSamuelson-randDist-p", p.lag, "-cooperation.rds", sep = "")))
  cat("\r")
}

randDistPlot = gridExtra::grid.arrange(grobs = list(randDistPlots[[1]], randDistPlots[[2]], 
                                                    randDistPlots[[3]], randDistPlots[[4]]), 
                                       widths = c(3, 3), heights = c(2, 2), nrow = 2)
ggsave(filename = here("Figures/AndreoniSamuelson-randDist_plots-cooperation.jpeg"), plot = randDistPlot)
