# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   The script analyzes the combined simulated results recorded in results/sim_results.rds
#   for the large N, size control simulations

library(tidyverse)

# Plot normal distribution results, N = 100 or N = 1000 --------------------------------------------
sim_res_normal <- readRDS("results/sim_results_largeN100.rds")
N.sims = 5000

# Null rejection rate
null_reject_rate_normal = sim_res_normal %>%
  mutate(
    pval_plag0 = 2*pmin(1-pnorm(plag0), pnorm(plag0)),
    pval_plag1 = 2*pmin(1-pnorm(plag1), pnorm(plag1)),
    reject_plag0 = ifelse(pval_plag0 <= 0.05, 1, 0), 
    reject_plag1 = ifelse(pval_plag1 <= 0.05, 1, 0) 
  )
null_reject_rate_normal = null_reject_rate_normal %>% 
  group_by(phi, pT) %>%
  summarise(reject_plag0 = sum(reject_plag0), 
            reject_plag1 = sum(reject_plag1)) %>%
  mutate(power_plag0 = reject_plag0/N.sims,
         power_plag1 = reject_plag1/N.sims) %>%
  select(-c(reject_plag0, reject_plag1))
null_reject_rate_normal = reshape2::melt(null_reject_rate_normal, id.vars = c("phi", "pT"))
levels(null_reject_rate_normal$variable) <- c("p = 0", "p = 1")
write.csv(null_reject_rate_normal, file = "Sim_Tables/NRR_N100_normal.csv")

# Histograms
p0_normal <- ggplot(sim_res_normal, aes(x = plag0, y = ..density..)) + 
  geom_histogram(bins = 50, fill = "#577ECB", boundary = 0) + 
  geom_vline(xintercept = 0, color = '#ED7D31', size = 1, linetype = "dashed") + 
  labs(x = "Z", y = "Density") + theme_bw()
plag0_plots_normal = p0_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag0_plots_normal, file = "Sim_Figures/UnitAvg_N100_normal_plag0_hist.png")

p1_normal <- ggplot(sim_res_normal, aes(x = plag1, y = ..density..)) + 
  geom_histogram(bins = 50, fill = "#577ECB", boundary = 0) + 
  geom_vline(xintercept = 0, color = '#ED7D31', size = 1, linetype = "dashed") + 
  labs(x = "Z", y = "Density") + theme_bw()
plag1_plots_normal = p1_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag1_plots_normal, file = "Sim_Figures/UnitAvg_N100_normal_plag1_hist.png", dpi = 150)

# QQ plots
p0_normal <- ggplot(sim_res_normal, aes(sample = plag0)) + 
  stat_qq(color = "#577ECB") + stat_qq_line(color = "#ED7D31") + theme_bw()
plag0_plots_normal = p0_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag0_plots_normal, file = "Sim_Figures/UnitAvg_N100_normal_plag0_qqplot.png", dpi = 150)

p1_normal <- ggplot(sim_res_normal, aes(sample = plag1)) + 
  stat_qq(color = "#577ECB") + stat_qq_line(color = "#ED7D31") + theme_bw()
plag1_plots_normal = p1_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag1_plots_normal, file = "Sim_Figures/UnitAvg_N100_normal_plag1_qqplot.png", dpi = 150)

# Plot cauchy distribution results, N = 50000 --------------------------------------------
sim_res_N50000_t <- readRDS("results/sim_results_largeN_t_normality.rds")

# Null rejection rate
null_reject_rate_N50000_t = sim_res_N50000_t %>%
  mutate(
    pval_plag0 = 2*pmin(1-pnorm(plag0), pnorm(plag0)),
    pval_plag1 = 2*pmin(1-pnorm(plag1), pnorm(plag1)),
    reject_plag0 = ifelse(pval_plag0 <= 0.05, 1, 0), 
    reject_plag1 = ifelse(pval_plag1 <= 0.05, 1, 0) 
  )
null_reject_rate_N50000_t = null_reject_rate_N50000_t %>% 
  group_by(phi, pT) %>%
  summarise(reject_plag0 = sum(reject_plag0), 
            reject_plag1 = sum(reject_plag1)) %>%
  mutate(power_plag0 = reject_plag0/N.sims,
         power_plag1 = reject_plag1/N.sims) %>%
  select(-c(reject_plag0, reject_plag1))
null_reject_rate_N50000_t = reshape2::melt(null_reject_rate_N50000_t, id.vars = c("phi", "pT"))
levels(null_reject_rate_N50000_t$variable) <- c("p = 0", "p = 1")
write.csv(null_reject_rate_N50000_t, file = "Sim_Tables/NRR_N50000_cauchy.csv")

# Histograms
p0_normal <- ggplot(sim_res_N50000_t, aes(x = plag0, y = ..density..)) + 
  geom_histogram(bins = 50, fill = "#577ECB", boundary = 0) + 
  geom_vline(xintercept = 0, color = '#ED7D31', size = 1, linetype = "dashed") + 
  labs(x = "Z", y = "Density") + theme_bw()
plag0_plots_normal = p0_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag0_plots_normal, file = "Sim_Figures/UnitAvg_N50000_t_plag0_hist.png", dpi = 150)

p1_normal <- ggplot(sim_res_N50000_t, aes(x = plag1, y = ..density..)) + 
  geom_histogram(bins = 50, fill = "#577ECB", boundary = 0) + 
  geom_vline(xintercept = 0, color = '#ED7D31', size = 1, linetype = "dashed") + 
  labs(x = "Z", y = "Density") + theme_bw()
plag1_plots_normal = p1_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag1_plots_normal, file = "Sim_Figures/UnitAvg_N50000_t_plag1_hist.png", dpi = 150)

# QQ plots
p0_normal <- ggplot(sim_res_N50000_t, aes(sample = plag0)) + 
  stat_qq(color = "#577ECB") + stat_qq_line(color = "#ED7D31") + theme_bw()
plag0_plots_normal = p0_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag0_plots_normal, file = "Sim_Figures/UnitAvg_N50000_t_plag0_qqplot.png", dpi = 150)

p1_normal <- ggplot(sim_res_N50000_t, aes(sample = plag1)) + 
  stat_qq(color = "#577ECB") + stat_qq_line(color = "#ED7D31") + theme_bw()
plag1_plots_normal = p1_normal + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(plag1_plots_normal, file = "Sim_Figures/UnitAvg_N50000_t_plag1_qqplot.png", dpi = 150)