# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   The script analyzes the combined simulated results recorded in results/sim_results.rds
#   for the large N, power simulations

library(tidyverse)
sim_res <- readRDS("results/sim_results_largeN_powerN.rds")
N.sims = 5000

# Compute power curves --------------------------------------------------------
sim_res = sim_res %>%
  mutate(
    pval_plag0 = 2*pmin(1-pnorm(plag0), pnorm(plag0)),
    pval_plag1 = 2*pmin(1-pnorm(plag1), pnorm(plag1)),
    reject_plag0 = ifelse(pval_plag0 <= 0.05, 1, 0), 
    reject_plag1 = ifelse(pval_plag1 <= 0.05, 1, 0) 
  )
power_res = sim_res %>% group_by(phi, beta, pT) %>%
  summarise(reject_plag0 = sum(reject_plag0), 
            reject_plag1 = sum(reject_plag1)) %>%
  mutate(power_plag0 = reject_plag0/N.sims,
         power_plag1 = reject_plag1/N.sims) %>%
  select(-c(reject_plag0, reject_plag1))
power_res = reshape2::melt(power_res, id.vars = c("phi", "beta", "pT"))
levels(power_res$variable) <- c("p = 0", "p = 1")

power = ggplot(power_res, aes(x = beta, y = value)) + 
  geom_line(aes(colour = variable)) +
  geom_point(aes(color = variable, shape = variable)) + 
  scale_colour_manual(values=c("#577ECB", "#ED7D31")) + 
  labs(x = latex2exp::TeX("$\\tau$"), y = "Rejection rate") + 
  theme_bw() + theme(legend.position = "none")
power_plots = power + facet_grid(rows = vars(phi), cols = vars(pT))
ggsave(power_plots, file = "Sim_Figures/LargeN/power_plots_largeN.png", dpi = 150)

# Compute ANOVA -------------
sim_res$pT = factor(sim_res$pT)
sim_res$phi = factor(sim_res$phi)
sim_res$beta = factor(sim_res$beta)

# Compute linear model for plag0
ANOVA.model.plag0 = lm(reject_plag0 ~ -1 + pT + phi + beta, data = sim_res)
var.plag0 = var(sim_res$reject_plag0)
var.pT.plag0 = var(ANOVA.model.plag0$coefficients[grep("pT", names(ANOVA.model.plag0$coefficients))])
var.phi.plag0 = var(ANOVA.model.plag0$coefficients[grep("phi", names(ANOVA.model.plag0$coefficients))])
var.beta.plag0 = var(ANOVA.model.plag0$coefficients[grep("beta", names(ANOVA.model.plag0$coefficients))])

cat("ANOVA for plag = 0", "\n")
cat("Fraction of var. associated w/ pT: ", 100*var.pT.plag0/var.plag0, "\n")
cat("Fraction of var. associated w/ phi: ", 100*var.phi.plag0/var.plag0, "\n")
cat("Fraction of var. associated w/ beta: ", 100*var.beta.plag0/var.plag0, "\n")

# Compute linear model for plag1
ANOVA.model.plag1 = lm(reject_plag1 ~ -1 + pT + phi + beta, data = sim_res)
var.plag1 = var(sim_res$reject_plag1)
var.pT.plag1 = var(ANOVA.model.plag1$coefficients[grep("pT", names(ANOVA.model.plag1$coefficients))])
var.phi.plag1 = var(ANOVA.model.plag1$coefficients[grep("phi", names(ANOVA.model.plag1$coefficients))])
var.beta.plag1 = var(ANOVA.model.plag1$coefficients[grep("beta", names(ANOVA.model.plag1$coefficients))])

cat("ANOVA for plag = 1", "\n")
cat("Fraction of var. associated w/ pT: ", 100*var.pT.plag1/var.plag1, "\n")
cat("Fraction of var. associated w/ phi: ", 100*var.phi.plag1/var.plag1, "\n")
cat("Fraction of var. associated w/ beta: ", 100*var.beta.plag1/var.plag1, "\n")