# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   Functions for estimating dynamic causal effects in randomized experiments 
#   on panel and time series data. 
#   
#   The functions here build upon the functions in "ts_analysis_functions.R" from the replication materials 
#   of Bojinov & Shephard (2019), which implement the q-step, lag-p causal effects (pg 9). Throughout, we 
#  set q.step = 0 to make this equivalent to the weighted-average i,t-th lag-p dynamic causal effect (definition 6) of
#  Bojinov et al. 
#
# Notes:
#   - We assume the data come in the form of a dataframe with a column containing the unit ids. 
#     That is, there is a variable that provides a unique identifier for each unit in the data.
#   - Helper functions that do not operate on this dataframe structure are masked from the 
#     user.
#   - We parallelize operations that are repeated across units. We use 
#     doParallel + foreach to easily implement this parallelization.

# Libraries -----------------------------------------------------------
library(furrr)
library(tidyverse)

# Variance estimator helper functions ---------------------------------
.cons_var_bound <- function(Yobs, Wobs, pT, p.lag, q.step  = 0) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes the conservative variance bound for the p-lag, q-step HT estimator. 
  #   This function is copied from the ts_analysis_function.R in Bojinov & Shephard (2019).
  #
  #   Notes: 
  #     1) We modify this function so that pT only accepts a vector of treatment probabilities.
  #     2) This is given in Eqn (5) of Bojinov Et Al in Section 3.2, which provides the expression for the i,t-th lag-p
  #        HT estimator. It is also given in Lemma 2 in Section B.2 of the online appendix of Bojinov & Shepard (2019).
  #     3) We mask this function from the user as it is used as a helper function
  #        to the functions defined below.
  #
  # Args:
  #   Yobs   = A vector of length T, the observed outcome path.
  #   Wobs   = A vector of treatment assignments
  #   pT     = A vector containing the probability of assigning a unit to treatment.
  #   p.lag  = A non-negative integer indicating the p-lag.
  #   q.step = A positive integer; indicating the number of steps. This is always set to zero.
  # Returns: List containing elements
  #   var.est   =  The estimated variance bound at each time point.
  #   gm.sq.est = The estimate of gm (the average of the variances). 
  
  timePeriods <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  K.step <- p.lag + q.step + 1
  if (K.step > 1) {
    pT.mat <- matrix(NA, nrow = K.step, ncol = timePeriods)
    pT.mat[1, ] <- pT * Wobs + (1 - pT) * (1 - Wobs)
    for (kk in 1:(K.step - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT.mat[1, 1:(timePeriods - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT * Wobs + (1 - pT) * (1 - Wobs)
  }
  var.est <- as.numeric(Yobs^2  * (1 + 2 * pT.prod * (2^(K.step - 1) - 1)) *
                          ((1 / 4)^(K.step - 1)) / pT.prod^2) 
  gm.sq.est <- sum(var.est[K.step:timePeriods]) / (timePeriods - K.step + 1) ^ 2
  return(list(var.est = var.est, gm.sq.est = gm.sq.est))
}

.var_fisher <- function(Yobs, Wobs, pT, p.lag, q.step = 0) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes the conservative variance bound for the p-lag, q-step HT estimator. 
  #   This function is copied from the ts_analysis_function.R in Bojinov & Shephard (2019).
  #
  #   Notes:
  #     1) We modify this function so that pT only accepts a vector of treatment probabilities.
  #     2) We mask this function from the user as it is used as a helper function
  #        to the functions defined below.
  #
  # Args:
  #   Yobs   = A vector of length T, the observed outcome path.
  #   Wobs   = A vector of treatment assignments
  #   pT     = A numeric vector of the probability of assigning a unit to treatment in each period
  #   p.lag  = A non-negative integer indicating the p-lag.
  #   q.step = A positive integer; indicating the number of steps. This is always set to zero.
  # 
  # Returns: List containing elements
  #   var.est   = The estimated variance bound at each time point.
  #   gm.sq.est = The estimate of gm (the average of the variances). 
  
  timePeriods <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  K.step <- p.lag + q.step
  if (K.step > 0) {
    binmat <- sapply(1:(2^(K.step + 1)), 
                     function(xx) {as.integer(intToBits(xx))[1:(K.step + 1)]})
    pinv <- sapply(1:ncol(binmat), function(xx) {1 / .pTvec(pT, binmat[, xx])})
    pT.prod <- rowSums(pinv)
  } else {
    pT.prod <- pT
  }
  if (K.step == 0){
    var.est <- as.numeric(Yobs^2 / (pT * (1 - pT)))
  } else {
    var.est <- as.numeric(Yobs^2  * ((1 / 4)^(K.step)) * pT.prod)
  }
  gm.sq.est <- sum(var.est[K.step:timePeriods]) / (timePeriods - K.step)^2
  return(list(var.est = var.est, gm.sq.est = gm.sq.est))
}

.pTvec <- function(pT, treat.vec) {
  # Description:
  #   This function is used in the computation of the Fisher variance. 
  # Args:
  #   pT        = The probability of receiving treatment at time T
  #   treat.vec = The treatment assignment vector
  # Returns:
  #   pT.out = A vector of length = length(pT) that gives the probability of
  #           observing the treat assignment specified by treat.vec at time t.
  timePeriods <- length(pT)
  k.step <- length(treat.vec)
  pT.mat <- matrix(1, nrow = k.step, ncol = timePeriods)
  pT.mat[1, ] <- c(treat.vec, rep(tail(treat.vec, 1), timePeriods - k.step)) * pT +
    (1 - c(treat.vec, rep(tail(treat.vec, 1), timePeriods - k.step))) * (1 - pT)
  if (k.step > 2) {
    for (jj in 2:(k.step - 1)) {
      Z <- c(treat.vec[1:(k.step - jj + 1)], 
             rep(tail(treat.vec[1:(k.step - jj + 1)], 1), timePeriods - k.step))
      pT.mat[jj, jj:timePeriods] <- 
        Z * pT[1:(timePeriods - jj + 1)] + (1 - Z) * (1 - pT[1:(timePeriods - jj + 1)])
    }
  }
  Z <- rep(treat.vec[1], timePeriods - k.step + 1)
  pT.mat[k.step, k.step:timePeriods] <- 
    (1 - Z) * (1 - pT[1:(timePeriods - k.step + 1)]) + Z * pT[1:(timePeriods - k.step + 1)]
  pT.out <- apply(pT.mat, 2, prod)
  return(pT.out)
}

# Horvitz Thompson estimator helper function --------------------------
.est_plag_qstep_ht <- function(Yobs, Wobs, pT, p.lag, q.step = 0, normalize, pT.treatmentprob) {
  # Description:
  #   This function computes the p-lag q-step estimator of the p-lag q-step 
  #   causal effect. p.lag = 0 corresponds to the contemporaneous causal effect.
  #
  #   This function is copied from the ts_analysis_function.R in Bojinov & Shephard (2019).
  #   It is used throughout the more general functions defined later on. 
  #
  #   Note: We modify this function so that pT only accepts a vector of treatment probabilities.
  #
  #   Note: We mask this function from the user as its outputs can be generated using the function
  #         unit.ht.wavg.est() defined below.
  #
  # Args:
  #   Yobs      = The observed data.
  #   Wobs      = The corresponding treatment assignments.
  #   pT        = A vector containing the probability of receiving treatment at each time poit.
  #   p.lag     = A non-negative integer indicating the p-lag.
  #   q.step    = A positive integer; indicating the number of steps. This is always set to zero.
  #   normalize = A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # Returns: List containing elements
  #   est.mean = The mean of the estimated values. 
  #   est      = The estimated value
  
  stopifnot((q.step + p.lag + 1) > 0)
  p.lag <- p.lag + 1
  T <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  if (pT.treatmentprob) {
    pT <- pT * Wobs + (1 - pT) * (1 - Wobs)
  }
  # Compute the probability of the received treatment
  if ((q.step + p.lag) > 1) {
    pT.mat <- matrix(NA, nrow = (q.step + p.lag), ncol = T)
    pT.mat[1, ] <- pT
    for (kk in 1:(q.step + p.lag - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT[1 : (T - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT
  }
  if (p.lag > 0){
    Wobs <- c(rep(0, p.lag - 1), Wobs[1:(T - p.lag + 1)])
  }
  est <- as.numeric(Yobs * ((-1)^(1 - Wobs)) * ((1 / 2) ^ ((p.lag + q.step) - 1)) / pT.prod)
  if (normalize) {
    gm <- .var_fisher(Yobs, Wobs, pT, p.lag, q.step)$var.est
    # Sometimes with very low probability gm = 0, then you just add some noise 
    # since 0/a = 0, this doesn't change anything 
    gm[which(gm == 0)] <- abs(jitter(gm[which(gm == 0)]))
    est <- est / sqrt(gm)
  }
  est.mean <- mean(est[(p.lag + q.step):T], na.rm = TRUE)
  return(list(est.mean = est.mean, est = est))
}

# Variance estimator functions ----------------------------------------
.unit_time_cons_var_bound <- function(expDF, unit, time, p.lag = 0) {
  # Description:
  #   This helper function provides is a wrapper to compute the conservative 
  #   variance bound for the weighted average i,t-th lag-p dynamic causal 
  #   effect at a user-specified unit and time period.
  #
  #   It allows the user to access the function .cons_var_bound() 
  #   by passing a dataframe of the entire experiment and unit id.
  #
  #   It is used by the function time.cons.var.bound().
  #
  # Inputs: 
  #   expDF = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit  = id of unit that we are interested in computing the HT estimator for.
  #   time  = time period
  #   p.lag = a non-negative integer indicating the p-lag.
  #
  # Outputs: List containing elements
  #   var.est   =  The estimated variance bound at each time point.
  #   gm.sq.est = The estimate of gm (the average of the variances).
  
  # Extract unit-level observations from expDF
  unitDF = expDF[expDF$unitId == unit, ]
  
  # Compute variance bound and return
  cons.var = .cons_var_bound(Yobs = unitDF$Yobs, Wobs = unitDF$Wobs, pT = unitDF$pT, p.lag = p.lag)$var.est
  return(cons.var[time])
}

unit.cons.var.bound <- function(expDF, unit, p.lag = 0) {
  # Description:
  #   This function provides is a wrapper to compute the conservative 
  #   variance bound for the weighted average i,t-th lag-p dynamic causal 
  #   effect.
  #
  #   It allows the user to access the function .cons_var_bound() 
  #   by passing a dataframe of the entire experiment and unit id.
  #
  # Inputs: 
  #   expDF = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit  = id of unit that we are interested in computing the HT estimator for.
  #   p.lag = a non-negative integer indicating the p-lag.
  #
  # Outputs: List containing elements
  #   var.est   =  The estimated variance bound at each time point.
  #   gm.sq.est = The estimate of gm (the average of the variances).
  
  # check if unit not a valid unit identifier.
  stopifnot(unit %in% expDF$unitId) 
  
  # Extract unit-level observations from expDF
  unitDF = expDF[expDF$unitId == unit, ]
  
  # Compute variance bound and return
  cons.var = .cons_var_bound(Yobs = unitDF$Yobs, Wobs = unitDF$Wobs, pT = unitDF$pT, p.lag = p.lag)
  return(list(var.est = cons.var$var.est, 
              gm.sq.est = cons.var$gm.sq.est))
}

time.cons.var.bound <- function(expDF, time, p.lag = 0, parallel = FALSE) {
  # Description:
  #   This function computes thec conservative variance bound
  #   for the cross-sectional averaged, time-specific weighted average i,t-th lag-p dynamic causal 
  #   effect.
  #
  # Note: This uses Theorem 3.2 in Bojinov et al and Equation (18)
  #
  # Inputs: 
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   time     = integer value specifying the time
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   parallel = logical flag for whether user would like to parallelize this operation.
  #
  # Outputs: Returns scalar containing the estimate of the variance.
  
  # Get vector of unit identifiers
  unitIds = unique(expDF$unitId)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # Type checking for valid inputs
  stopifnot(time %% 1 == 0, time > 0, time >= p.lag + 1, time <= timePeriods)
  
  # Compute the conservative variance estimate for each unit and collect them for the time period of interest
  # Loop through units, compute HT estimator, store HT estimate associated with user-specified period
  if (parallel) {
    plan(multiprocess)
    unit.var.bound = future_map_dbl(.x = unitIds, .f = ~.unit_time_cons_var_bound(expDF = expDF, unit = .x, 
                                                                                  time = time, p.lag = p.lag))
  } else {
    unit.var.bound = map_dbl(.x = unitIds, .f = ~.unit_time_cons_var_bound(expDF = expDF, unit = .x, 
                                                                           time = time, p.lag = p.lag))
  }
  # Compute the cross-sectional average and return
  return( sum(unit.var.bound, na.rm = TRUE) / (length(unitIds) ^ 2) )
}

total.cons.var.bound <- function(expDF, p.lag = 0, parallel = FALSE) {
  # Description:
  #   This function computes thec conservative variance bound
  #   for the cross-sectional and time averaged, weighted average i,t-th lag-p dynamic causal 
  #   effect.
  #
  # Note: This uses Theorem 3.2 in Bojinov et al and Equation (18)
  #
  # Inputs: 
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   parallel = logical flag for whether user would like to parallelize this operation.
  #
  # Outputs: Returns scalar containing the estimate of the variance.
  
  # Get vector of unit identifiers
  unitIds = unique(expDF$unitId)
  N = length(unitIds)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # Loop through units, compute HT estimator, store HT estimate associated with each unit in N x T matrix
  if (parallel) {
    plan(multiprocess)
    unit.var.bound = future_map(.x = unitIds, .f = ~unit.cons.var.bound(expDF = expDF, unit = .x, p.lag = p.lag))
  } else {
    unit.var.bound = map(.x = unitIds, .f = ~unit.cons.var.bound(expDF = expDF, unit = .x, p.lag = p.lag))
  }
  unit.var.bound = sapply(unit.var.bound, `[[`, 1)
  return( sum(unit.var.bound[(p.lag+1):timePeriods,], na.rm = TRUE) / (N * (timePeriods-p.lag))^2 )
}


# Horvitz Thompson estimator functions --------------------------------
.unit_time_ht_wavg_est <- function(expDF, unit, time, p.lag = 0, normalize = FALSE, pT.treatmentprob = TRUE) {
  # Description: 
  #   This helper function computes the period-specific, unit-specific horvitz thompson
  #   estimator of the weighted averaged i,t-th lag-p dynamic causal effect, where we
  #   we place equal weights on all paths v in W^{p}.
  #
  #   It returns the HT estimator for a user-specified time period AND unit. This function
  #   is used as a helper function in time.ht.wavg.est()
  #
  #   Note: This function uses .est_plag_qstep_ht() to construct the weighted-average i,t-th lag-p,
  #         q-step causal effect
  #
  # Inputs:
  #   expDF = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit  = id of unit that we are interested in computing the HT estimator for.
  #   time  = time period of interest
  #   p.lag = a non-negative integer indicating the p-lag.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #             
  # Outputs: Scalar that contains the HT estimate for the chosen unit and time period
  
  # Extract unit-level observations from expDF
  unitDF = expDF[expDF$unitId == unit, ]
  
  # Compute the weighted-average i,t-th lag-p causal effect
  # and its average across time periods
  plag = .est_plag_qstep_ht(Yobs = unitDF$Yobs, Wobs = unitDF$Wobs, 
                            pT = unitDF$pT, p.lag = p.lag, normalize = normalize, pT.treatmentprob = pT.treatmentprob)$est
  return(plag[time])
}

unit.ht.wavg.est <- function(expDF, unit, p.lag = 0, normalize = FALSE, pT.treatmentprob = TRUE) {
  # Description: 
  #   This function computes period-specific and unit-specific horvitz thompson
  #   estimator of the weighted averaged i,t-th lag-p dynamic causal effect, where we
  #   we place equal weights on all paths v in W^{p}.
  #
  #   The weighted average i,t-th lag-p dyanmic causal effect
  #   corresponds to the generalized weighted average i, t-th lag-p dynamic causal effect
  #   defined in Section 2.3.1 of in Bojinov Et Al.
  #
  #   The time average of the weighted average i,t-th lag-p dynamic causal effect
  #   simply averages this over time periods.
  #
  #   Note: This function uses .est_plag_qstep_ht() to construct the weighted-average i,t-th lag-p,
  #         q-step causal effect
  #
  # Inputs:
  #   expDF = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit  = id of unit that we are interested in computing the HT estimator for.
  #   p.lag = a non-negative integer indicating the p-lag.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #             
  # Outputs: list that contains
  #    1) unit.est      = period-specific HT estimates for the chosen unit
  #    2) unit.tavg.est = time average of the unit specific HT estimates
  
  # check if unit not a valid unit identifier.
  stopifnot(unit %in% expDF$unitId) 
  
  # Extract unit-level observations from expDF
  unitDF = expDF[expDF$unitId == unit, ]
  
  # Compute the weighted-average i,t-th lag-p, dynamic causal effect
  # and its average across time periods
  plag = .est_plag_qstep_ht(Yobs = unitDF$Yobs, Wobs = unitDF$Wobs, pT = unitDF$pT,
                            p.lag = p.lag, normalize = normalize, pT.treatmentprob = pT.treatmentprob)
  return(list(unit.est = plag$est, unit.tavg.est = plag$est.mean))
}

time.ht.wavg.est <- function(expDF, time, p.lag = 0, normalize = FALSE, pT.treatmentprob = TRUE, parallel = FALSE) {
  # Description: 
  #   This function computes the period-specific cross-sectional average of
  #   the weighted-average i,t-th lag-p dynamic causal effect, where we place equal weights on 
  #   all paths v \in W^{p}.
  #
  # Inputs:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   time     = integer value specifying the time
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   parallel = logical flag for whether user would like to parallelize this operation.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Outputs: returns double that equals the cross-sectional average of HT estimates for the chosen time period
  
  # Get vector of unit identifiers
  unitIds = unique(expDF$unitId)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # Type checking for valid inputs
  stopifnot(time %% 1 == 0, time > 0, time >= p.lag + 1, time <= timePeriods)
  
  # Loop through units, compute HT estimator, store HT estimate associated with user-specified period
  if (parallel) {
    plan(multiprocess)
    unit.ht = future_map_dbl(.x = unitIds, .f = ~.unit_time_ht_wavg_est(expDF = expDF, unit = .x, 
                                                                        time = time, p.lag = p.lag,
                                                                        pT.treatmentprob = pT.treatmentprob, normalize = normalize))
  } else {
    unit.ht = map_dbl(.x = unitIds, .f = ~.unit_time_ht_wavg_est(expDF = expDF, unit = .x, 
                                                                 time = time, p.lag = p.lag, 
                                                                 pT.treatmentprob = pT.treatmentprob, normalize = normalize))
  }
  return(sum(unit.ht, na.rm = TRUE) / length(unitIds))
}

total.ht.wavg.est <- function(expDF, p.lag = 0, normalize = FALSE, pT.treatmentprob = TRUE, parallel = FALSE) {
  # Description:
  #   This function computes grand average of the unit-specific weighted-average i,t-th lag-p dynamic causal effect, 
  #   where we place equal weights on all paths v \in W^{p}. 
  # 
  # Inputs:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance.
  #   parallel = logical flag for whether user would like to parallelize this operation.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #
  # Outputs: returns double that equals the grand average of HT estimates
  
  # Get vector of unit identifiers
  unitIds = unique(expDF$unitId)
  N = length(unitIds)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # Loop through units, compute HT estimator, store HT estimate associated with each unit in N x T matrix
  if (parallel) {
    plan(multiprocess)
    unit.ht = future_map(.x = unitIds, .f = ~unit.ht.wavg.est(expDF = expDF, unit = .x, 
                                                              p.lag = p.lag, pT.treatmentprob = pT.treatmentprob,
                                                              normalize = normalize))
  } else {
    unit.ht = map(.x = unitIds, .f = ~unit.ht.wavg.est(expDF = expDF, unit = .x, 
                                                       p.lag = p.lag, pT.treatmentprob = pT.treatmentprob,
                                                       normalize = normalize))
  }
  # Compute the grand average and return
  unit.ht = sapply(unit.ht, `[[`, 1)
  return(mean(unit.ht[(p.lag+1):timePeriods,], na.rm = TRUE))
}

# Asymptotic conservative inference functions -------------------------
unit.normal.pval <- function(expDF, unit, p.lag = 0, returnCI = FALSE, alpha = 0.05, pT.treatmentprob = TRUE) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes conservative p-value as well as a (1-ap)% confidence interval 
  #   (whenever rCI == TRUE) for the time-average causal effect of a user-specified unit. 
  #
  # Note: This uses results in Bojinov & Shephard (2019) and Theorem 3.2 in Bojinov et al.
  # 
  # Args:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #              2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #              4) pT -- period-specific treatment probabilities for each unit.
  #   unit     = id of unit that we are interested in computing the HT estimator for.
  #   p.lag    = A non-negative integer indicating the p-lag.
  #   returnCI = A logical value indicating if a 95% normal CI should be returned
  #   alpha    = Significance level for the confidence interval.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Returns:
  #   est = time-averaged, unit specific HT estimator
  #   Z: The Z statistic.
  #   p.val: The estimated p-value.
  #   CI = confidence interval if returnCI = TRUE
  
  # Time periods
  timePeriods = nrow(expDF[expDF$unitId == unit, ])
  
  # Compute time-averaged, unit specific HT estimator
  est = unit.ht.wavg.est(expDF = expDF, unit = unit, p.lag = p.lag, normalize = FALSE, pT.treatmentprob = pT.treatmentprob)$unit.tavg.est
  
  # Compute conservative variance of time-averaged, unit specific HT estimator
  sigma = sqrt(unit.cons.var.bound(expDF = expDF, unit = unit, p.lag = p.lag)$gm.sq.est)
  
  # Compute z-statistic and p-value
  Z <- est / sigma
  p.val <- 2 * min(1 - pnorm(Z), pnorm(Z))
  
  # If returnCI, construct confidence interval
  if (returnCI) {
    CI <- est + c(sigma * qnorm(1 - alpha / 2), - sigma * qnorm(1 - alpha / 2))
    out <- list(unit.tavg.est = est, Z = Z, p.val = p.val, CI = CI)
  } else {
    out <- list(unit.tavg.est = est, Z = Z, p.val = p.val)
  }
  return(out)
}


time.normal.pval <- function(expDF, time, p.lag = 0, returnCI = FALSE, alpha = 0.05, pT.treatmentprob = TRUE) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes conservative p-value as well as a (1-alpha)% confidence interval 
  #   (whenever rCI == TRUE) for the cross-sectional average causal effect of a user-specified time period. 
  #
  # Note: This uses results in Bojinov & Shephard (2019) and Theorem 3.2 in Bojinov et al.
  # 
  # Args:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #              2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #              4) pT -- period-specific treatment probabilities for each unit.
  #   time     = integer value specifying the time
  #   p.lag    = A non-negative integer indicating the p-lag.
  #   returnCI = A logical value indicating if a 95% normal CI should be returned
  #   alpha    = Significance level for the confidence interval.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Returns:
  #   est = time-averaged, unit specific HT estimator
  #   Z: The Z statistic.
  #   p.val: The estimated p-value.
  #   CI = confidence interval if returnCI = TRUE
  
  # compute cross-sectional average HT estimator
  cross.wavg.est = time.ht.wavg.est(expDF = expDF, time = time, p.lag = p.lag, pT.treatmentprob = pT.treatmentprob)
  
  # Compute conservative variance of time-averaged, unit specific HT estimator
  sigma = sqrt(time.cons.var.bound(expDF = expDF, time = time, p.lag = p.lag))
  
  # Compute z-statistic and p-value
  Z <- cross.wavg.est / sigma
  p.val <- 2 * min(1 - pnorm(Z), pnorm(Z))
  
  # If returnCI, construct confidence interval
  if (returnCI) {
    CI <- cross.wavg.est + c(sigma * qnorm(1 - alpha / 2), - sigma * qnorm(1 - alpha / 2))
    out <- list(cross.wavg.est = cross.wavg.est, Z = Z, p.val = p.val, CI = CI)
  } else {
    out <- list(cross.wavg.est = cross.wavg.est, Z = Z, p.val = p.val)
  }
  return(out)
}

total.normal.pval <- function(expDF, p.lag = 0, returnCI = FALSE, alpha = 0.05, pT.treatmentprob = TRUE) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes conservative p-value as well as a (1-alpha)% confidence interval 
  #   (whenever rCI == TRUE) for the cross-sectional and time averaged causal effect. 
  #
  # Note: This uses results in Bojinov & Shephard (2019) and Theorem 3.2 in Bojinov et al.
  # 
  # Args:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #              2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #              4) pT -- period-specific treatment probabilities for each unit.
  #   p.lag    = A non-negative integer indicating the p-lag.
  #   returnCI = A logical value indicating if a 95% normal CI should be returned
  #   alpha    = Significance level for the confidence interval.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Returns:
  #   est = time-averaged, unit specific HT estimator
  #   Z: The Z statistic.
  #   p.val: The estimated p-value.
  #   CI = confidence interval if returnCI = TRUE
  
  # compute total average HT estimator
  total.wavg.est = total.ht.wavg.est(expDF = expDF, p.lag = p.lag, pT.treatmentprob = pT.treatmentprob)
  
  # Compute conservative variance of time-averaged, unit specific HT estimator
  sigma = sqrt(total.cons.var.bound(expDF = expDF, p.lag = p.lag))
  
  # Compute z-statistic and p-value
  Z <- total.wavg.est / sigma
  p.val <- 2 * min(1 - pnorm(Z), pnorm(Z))
  
  # If returnCI, construct confidence interval
  if (returnCI) {
    CI <- total.wavg.est + c(sigma * qnorm(1 - alpha / 2), - sigma * qnorm(1 - alpha / 2))
    out <- list(total.wavg.est = total.wavg.est, Z = Z, p.val = p.val, CI = CI)
  } else {
    out <- list(total.wavg.est = total.wavg.est, Z = Z, p.val = p.val)
  }
  return(out)
}

# Randomization p-value helper functions ------------------------------
.redraw_unit_time_rand_est <- function(unit, time, expDF, p.lag) {
  # Description: This is a helper function that constructs a draw
  # from the randomization distribution for a specified unit, time.
  # It redraws the treatments from period time-p.lag to period time.
  unitDF = expDF[expDF$unitId == unit, ]
  unitDF$Wobs[(time-p.lag):time] = rbinom(p.lag+1, 1, prob = unitDF$pT[(time-p.lag):time])
  return(unitDF)
}

.compute_unit_rand_est <- function(sim, unitDF, unit, p.lag, null.val, pT.treatmentprob) {
  # Description:
  #   This function constructs a draw from the randomization distribution, 
  #   computes the value of the time-averaged weighted lag-p dynamic causal 
  #   effect for a user-specified unit and returns that value. 
  #
  #  This function is used as a helper to unit.rand.pval()
  #
  # Inputs:
  #   sim      = simulation number -- dummy input used by the function map().
  #   unitDF    = N*T x 4 dataframe that contains columns 
  #               1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 
  #               3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit     = id of unit that we are interested in computing the HT estimator for.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #
  # Output: Scalar containing the estimate under the randomization distribution
  unitDF$Wobs = rbinom(nrow(unitDF), 1, prob = unitDF$pT)
  return(unit.ht.wavg.est(expDF = unitDF, unit = unit, p.lag = p.lag, pT.treatmentprob = pT.treatmentprob)$unit.tavg.est + null.val)
}

.compute_time_rand_est <- function(sim, expDF, time, unitIds, p.lag, null.val, pT.treatmentprob) {
  # Description:
  #   This function constructs a draw from the randomization distribution, 
  #   computes the value of the cross-sectional averaged weighted lag-p dynamic causal 
  #   effect for a user-specified time and returns that value. 
  #
  #  This function is used as a helper to time.rand.pval()
  #
  # Inputs:
  #   sim      = simulation number -- dummy input used by the function map().
  #   expDF    = N*T x 4 dataframe that contains columns 
  #               1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 
  #               3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   time     = time period of interest.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #
  # Output: Scalar containing the estimate under the randomization distribution
  
  # We re-draw the treatments from period t-p to period t for each unit, then re-compute the estimate.
  expDFnew = map_dfr(.x = unitIds, .f = ~.redraw_unit_time_rand_est(unit = .x, time = time, expDF = expDF, p.lag = p.lag))
  return(time.ht.wavg.est(expDF = expDFnew, time = time, p.lag = p.lag, 
                          pT.treatmentprob = pT.treatmentprob, parallel = FALSE) + null.val)
}

.compute_total_rand_est <- function(sim, expDF, unitIds, timePeriods, p.lag, null.val, pT.treatmentprob) {
  # Description:
  #   This function constructs a draw from the randomization distribution, 
  #   computes the value of the cgrand averaged weighted lag-p dynamic causal 
  #   effect for a user-specified time and returns that value. 
  #
  #  This function is used as a helper to time.rand.pval()
  #
  # Inputs:
  #   sim      = simulation number -- dummy input used by the function map().
  #   expDF    = N*T x 4 dataframe that contains columns 
  #               1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 
  #               3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #
  # Output: Scalar containing the estimate under the randomization distribution
  cs_ests = map_dbl(.x = (p.lag+1):timePeriods, 
                    .f = ~.compute_time_rand_est(sim = 1, expDF = expDF, 
                                                 time = .x, unitIds = unitIds, 
                                                 p.lag = p.lag, pT.treatmentprob = pT.treatmentprob, 
                                                 null.val = null.val))
  return(mean(cs_ests, na.rm = TRUE))
}


# Randomization p-values functions ------------------------------------
unit.rand.pval <- function(expDF, unit, p.lag = 0, null.val = 0, N.sims = 1000, parallel = FALSE, pT.treatmentprob = TRUE) {
  # Description:
  #  This functions takes an observed outcome path and treatment path for a user-specified unit, and
  #  computes a randomization based p-value under the sharp null that the causal effects are time-invariant
  #  for a user-specified unit.
  # 
  # Inputs:
  #   expDF    = T x 4 dataframe that contains columns 
  #               1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 
  #               3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   unit     = id of unit that we are interested in computing the HT estimator for.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   N.sims   = An integer that specifies the number of randomizations used to generate the p-value.
  #   parallel = logical flag to parallelize this operation.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Outputs: returns a list containing
  #   1) est       = initial estimate
  #   2) rand.pval = randomization p-value
  #   3) rand.dist = vector containing randomization distribution
  
  # Compute initial estimate
  init.est = unit.ht.wavg.est(expDF = expDF, unit = unit, p.lag = p.lag, pT.treatmentprob = pT.treatmentprob)$unit.tavg.est
  unitDF = expDF[expDF$unitId == unit, ]
  
  # compute randomization distribution
  if (parallel) {
    plan(multiprocess)
    rand.dist = future_map_dbl(.x = 1:N.sims, .f = ~.compute_unit_rand_est(sim = .x, unitDF = unitDF, unit = unit,
                                                                           p.lag = p.lag, null.val = null.val,
                                                                           pT.treatmentprob = pT.treatmentprob))
  } else {
    rand.dist = map_dbl(.x = 1:N.sims, .f = ~.compute_unit_rand_est(sim = .x, unitDF = unitDF, unit = unit, 
                                                                    p.lag = p.lag, null.val = null.val, 
                                                                    pT.treatmentprob = pT.treatmentprob))
  }
  pval = sum(ifelse(abs(rand.dist) > abs(init.est), 1, 0))/N.sims # compute p value
  return(list(est = init.est, rand.pval = pval, rand.dist = rand.dist))
}

time.rand.pval <- function(expDF, time, p.lag = 0, null.val = 0, N.sims = 1000, parallel = FALSE, pT.treatmentprob = TRUE) {
  # Description:
  #  This functions takes an observed outcome paths and treatment paths, and
  #  computes a randomization based p-value under the sharp null that the causal effects are unit-invariant
  #  for a user-specified period. 
  #
  # Inputs:
  #   expDF    = N*T x 4 dataframe that contains columns 
  #               1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 
  #               3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   time     = integer value specifying the time
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   N.sims   = An integer that specifies the number of randomizations used to generate the p-value.
  #   parallel = logical flag to parallelize this operation.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # 
  # Outputs: returns a list containing
  #   1) est       = initial estimate
  #   2) rand.pval = randomization p-value
  #   3) rand.dist = vector containing randomization distribution
  
  # compute initial estimate
  init.est = time.ht.wavg.est(expDF = expDF, time = time, p.lag = p.lag, parallel = parallel, pT.treatmentprob = pT.treatmentprob)
  
  # extract size of panel, N and T
  unitIds = unique(expDF$unitId)
  N = length(unitIds)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # compute randomization distribution
  if (parallel) {
    plan(multiprocess)
    rand.dist = future_map_dbl(.x = 1:N.sims, .f = ~.compute_time_rand_est(sim = .x, expDF = expDF, 
                                                                           time = time, unitIds = unitIds, 
                                                                           p.lag = p.lag, null.val = null.val, 
                                                                           pT.treatmentprob = pT.treatmentprob))
  } else {
    rand.dist = map_dbl(.x = 1:N.sims, .f = ~.compute_time_rand_est(sim = .x, expDF = expDF, 
                                                                    time = time, unitIds = unitIds, 
                                                                    p.lag = p.lag, null.val = null.val, 
                                                                    pT.treatmentprob = pT.treatmentprob))
  }
  pval = sum(ifelse(abs(rand.dist) > abs(init.est), 1, 0))/N.sims # compute p value
  return(list(est = init.est, rand.pval = pval, rand.dist = rand.dist))
}

total.rand.pval <- function(expDF, p.lag = 0, null.val = 0, N.sims = 1000, parallel = FALSE, pT.treatmentprob = TRUE) {
  # Description:
  #  This functions takes an observed outcome paths and treatment paths, and
  #  computes a randomization based p-value under the sharp null that the causal effects are unit-invariant and
  #  time-invariant across the full panel. 
  # 
  # Inputs:
  #   expDF    = N*T x 4 dataframe that contains columns 1) unitId -- unique identifiers for units,
  #               2) Yobs -- observed outcomes, 3) Wobs -- observed treatments, 
  #               4) pT -- period-specific treatment probabilities for each unit.
  #   p.lag    = a non-negative integer indicating the p-lag.
  #   null.val = The null value we wish to test against, this is used to build CI
  #   N.sims   = An integer that specifies the number of randomizations used to generate the p-value.
  #   parallel = logical flag to parallelize this operation.
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  #
  # Outputs: returns a list containing
  #   1) est       = initial estimate
  #   2) rand.pval = randomization p-value
  #   3) rand.dist = vector containing randomization distribution
  
  # compute initial estimate
  init.est = total.ht.wavg.est(expDF = expDF, p.lag = p.lag, parallel = parallel)
  
  # extract size of panel, N and T
  unitIds = unique(expDF$unitId)
  N = length(unitIds)
  timePeriods = nrow(expDF[expDF$unitId == unitIds[1], ])
  
  # compute randomization distribution
  if (parallel) {
    plan(multiprocess)
    rand.dist = future_map_dbl(.x = 1:N.sims, .f = ~.compute_total_rand_est(sim = .x, expDF = expDF, 
                                                                            unitIds = unitIds, timePeriods = timePeriods, 
                                                                            p.lag = p.lag, null.val = null.val, 
                                                                            pT.treatmentprob = pT.treatmentprob))
  } else {
    rand.dist = map_dbl(.x = 1:N.sims, .f = ~.compute_total_rand_est(sim = .x, expDF = expDF, 
                                                                     unitIds = unitIds, timePeriods = timePeriods, 
                                                                     p.lag = p.lag, null.val = null.val, 
                                                                     pT.treatmentprob = pT.treatmentprob))
  }
  pval = sum(ifelse(abs(rand.dist) > abs(init.est), 1, 0))/N.sims # compute p value
  return(list(est = init.est, rand.pval = pval, rand.dist = rand.dist))
}