# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   Functions used for the simulation studies in Bojinov, Rambachan & Shephard (2020).

# Libraries -----------------------------------------------------------
library(tidyverse)
library(foreach)
library(parallel)
library(doParallel)

# Functions to generate data ------------------------------------------
AR1.unit.gen.obs <- function(timePeriods, phi, beta, pT, err.dist = "normal", sigma = 1, sigma.full = NULL) {
  # Description:
  #   This function generates an AR(1) time series for a given unit following the
  #   an autoregressive potential outcome panels. We
  #   model the AR(1) process as being driven by Gaussian innovations or
  #   student-t innovations.
  #
  #   This function imposes that the treatment only has a contemporaneous effect
  #   on the outcome of interest in the AR process: beta_t = 0 for all t \neq 0.
  #
  # Inputs:
  #   timePeriods= A positive integer indicating the total number of time steps.
  #   phi        = number specifying the AR coefficient.
  #   beta       = number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT         = T x 1 vector giving the treatment probabilities across periods.
  #   err.dist   = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma      = variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   sigma.full = T x 1 vector giving the errors. This is useful
  #               for generating the full science table - as the error terms
  #               are constant accross simulations
  #
  # Outputs: Returns dataframe with columns
  #   Yobs       = observed Y values
  #   Wobs       = observed treatment values
  #   sigma.full = vector containing simulated errors
  
  if (timePeriods <= 0 | timePeriods %% 1 != 0) {
    stop('T must be a positive integer')
  }
  
  Wobs = rbinom(timePeriods, 1, pT) # Draw observed treatments
  if (is.null(sigma.full)) { # if sigma.full is null, draw errors
    if (err.dist == 'normal') { # Draw errors if normal
      sigma.full = rnorm(timePeriods, 0, sigma)
    } else if (err.dist == 't') { # Draw errors if student t
      sigma.full = rt(timePeriods, df = sigma)
    } else { # return warning if incorrect input
      stop('err.dist must equal normal or t')
    }
  }
  x = beta*Wobs + sigma.full # Construct initial time series
  Yobs = as.numeric(stats::filter(x, phi, method = 'recursive')) # Construct Y by filtering x
  return(tibble(Yobs = Yobs, Wobs = Wobs, pT = pT, sigma.full = sigma.full))
}

AR1.panel.gen.obs <- function(N, timePeriods, phi, beta, pT, err.dist = "normal", sigma = 1, 
                              sigma.full = NULL, parallel = FALSE) {
  # Description:
  #   This function generates an AR(1) time series for a full panel following the
  #   expression in Example 1 on autoregressive potential outcome panels. We
  #   model the AR(1) process as being driven by Gaussian innovations, or
  #   student-t distribution.
  #
  #   This function imposes that the treatment only has a contemporaneous effect
  #   on the outcome of interest in the AR process: beta_t = 0 for all t \neq 0.
  # 
  # Inputs:
  #   N          = A positive integer indicating the total number of units
  #   timePeriods = A positive integer indicating the total number of time steps.
  #   phi        = N x 1 number specifying the AR coefficients for each unit
  #   beta       = N x 1 number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT         = N x T matrix giving the unit-specific treatment probabilities across periods.
  #   err.dist   = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma      = specifies the variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   sigma.full = N x T vector giving the errors. This is useful
  #               for generating the full science table - as the error terms
  #               are constant accross simulations
  #
  # Outputs: Returns a NT x 5 dataframe containing the columns
  #   unitId     = unit level identifiers
  #   Yobs       = observed outcomes
  #   Wobs       = observed treatments
  #   pT         = treatment probabilities
  #   sigma.full = simulated errors.
  
  if (N <= 0 | N %% 1 != 0) {
    stop('N must be a positive integer')
  }
  
  # Loop through units, generating observations. We parallelize this operation.
  if (parallel) {
    cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
    doParallel::registerDoParallel(cores-1)
    obs = foreach(i = 1:N, .combine = 'rbind') %dopar% {
      temp = AR1.unit.gen.obs(timePeriods = timePeriods, phi = phi[i], beta = beta[i], pT = pT[i, ], 
                              err.dist = err.dist, sigma = sigma, sigma.full = sigma.full[i,])
      temp$unitId = rep(i, timePeriods)
      temp
    }
  } else {
    obs = foreach(i = 1:N, .combine = 'rbind') %do% {
      temp = AR1.unit.gen.obs(timePeriods = timePeriods, phi = phi[i], beta = beta[i], pT = pT[i, ], 
                              err.dist = err.dist, sigma = sigma, sigma.full = sigma.full[i,])
      temp$unitId = rep(i, timePeriods)
      temp
    }
  }
  return(obs)
}

AR1.panel.gen.multiple_largeT <- function(sims, timePeriods, Nunits, phi, beta, pT, err.dist, sigma, 
                                          same.err, seed, parallel) {
  # Description
  #   This function computes multiple draws of a panel experiment 
  #   at a given choice of parameters using the same errors. This is
  #   useful for analyzing the properties of the estimator over the randomizaton distribution.
  #
  #   It returns a tibble that contains the unit-based p-lag = 0, 1, 2 estimates.
  # 
  # Args:
  #   sims        = A number indicating the number of paths of the science table to try.
  #   timePeriods = A number indicating the total number of time setps
  #   Nunits      = A number specifying the number of units
  #   phi         = N x 1 number specifying the AR coefficients for each unit
  #   beta        = N x 1 number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT          = N x T matrix giving the unit-specific treatment probabilities across periods.
  #   err.dist    = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma       = specifies the variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   seed        = seed
  #   parallel    = flag for parallel option with AR1.panel.gen.obs
  #
  # Returns: Returns a sims*(Nunits*timePeriods) x 5 dataframe containing the columns
  #   unitId     = unit level identifiers
  #   Yobs       = observed outcomes
  #   Wobs       = observed treatments
  #   pT         = treatment probabilities
  #   sim        = simulation number.
  
  # Check error dist
  stopifnot(err.dist == "normal" | err.dist == "t")
  
  # set seed
  set.seed(seed)
  
  # Draw errors
  if (err.dist == "normal") {
    sigma.full = matrix(rnorm(Nunits*timePeriods, mean = 0, sd = 1), nrow = Nunits, ncol = timePeriods)
  } else {
    sigma.full = matrix(rt(Nunits*timePeriods, df = 1), nrow = Nunits, ncol = timePeriods)
  }
  
  # Redraw AR1 panel with these simulation errors
  if (parallel) {
    cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
    doParallel::registerDoParallel(cores)
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %dopar% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full, parallel = FALSE)
      tibble(
        plag0 = unit.normal.pval(expDF = temp, unit = 1, p.lag = 0)$Z,
        plag1 = unit.normal.pval(expDF = temp, unit = 1, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  } else {
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %do% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full)
      tibble(
        plag0 = unit.normal.pval(expDF = temp, unit = 1, p.lag = 0)$Z,
        plag1 = unit.normal.pval(expDF = temp, unit = 1, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  }
  return(simulated_data)
}

AR1.panel.gen.multiple_largeN <- function(sims, timePeriods, Nunits, phi, beta, pT, err.dist, sigma, 
                                          same.err, seed, parallel) {
  # Description
  #   This function computes multiple draws of a panel experiment 
  #   at a given choice of parameters using the same errors. This is
  #   useful for analyzing the properties of the estimator over the randomizaton distribution.
  #
  #   It returns a tibble that contains the time-specific p-lag = 0, 1, 2 estimates in the last period
  # 
  # Args:
  #   sims        = A number indicating the number of paths of the science table to try.
  #   timePeriods = A number indicating the total number of time setps
  #   Nunits      = A number specifying the number of units
  #   phi         = N x 1 number specifying the AR coefficients for each unit
  #   beta        = N x 1 number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT          = N x T matrix giving the unit-specific treatment probabilities across periods.
  #   err.dist    = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma       = specifies the variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   seed        = seed
  #   parallel    = flag for parallel option with AR1.panel.gen.obs
  #
  # Returns: Returns a sims*(Nunits*timePeriods) x 5 dataframe containing the columns
  #   unitId     = unit level identifiers
  #   Yobs       = observed outcomes
  #   Wobs       = observed treatments
  #   pT         = treatment probabilities
  #   sim        = simulation number.
  
  # Check error dist
  stopifnot(err.dist == "normal" | err.dist == "t")
  
  # set seed
  set.seed(seed)
  
  # Draw errors
  if (err.dist == "normal") {
    sigma.full = matrix(rnorm(Nunits*timePeriods, mean = 0, sd = 1), nrow = Nunits, ncol = timePeriods)
  } else {
    sigma.full = matrix(rt(Nunits*timePeriods, df = 1), nrow = Nunits, ncol = timePeriods)
  }
  
  # Redraw AR1 panel with these simulation errors
  if (parallel) {
    cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
    doParallel::registerDoParallel(cores)
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %dopar% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full, parallel = FALSE)
      tibble(
        plag0 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 0)$Z,
        plag1 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  } else {
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %do% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full)
      tibble(
        plag0 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 0)$Z,
        plag1 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  }
  return(simulated_data)
}

AR1.panel.gen.multiple_largeNT <- function(sims, timePeriods, Nunits, phi, beta, pT, err.dist, sigma, 
                                           same.err, seed, parallel) {
  # Description
  #   This function computes multiple draws of a panel experiment 
  #   at a given choice of parameters using the same errors. This is
  #   useful for analyzing the properties of the estimator over the randomizaton distribution.
  #
  #   It returns a tibble that contains the total p-lag = 0, 1, 2 estimates in the last period
  # 
  # Args:
  #   sims        = A number indicating the number of paths of the science table to try.
  #   timePeriods = A number indicating the total number of time setps
  #   Nunits      = A number specifying the number of units
  #   phi         = N x 1 number specifying the AR coefficients for each unit
  #   beta        = N x 1 number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT          = N x T matrix giving the unit-specific treatment probabilities across periods.
  #   err.dist    = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma       = specifies the variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   seed        = seed
  #   parallel    = flag for parallel option with AR1.panel.gen.obs
  #
  # Returns: Returns a sims*(Nunits*timePeriods) x 5 dataframe containing the columns
  #   unitId     = unit level identifiers
  #   Yobs       = observed outcomes
  #   Wobs       = observed treatments
  #   pT         = treatment probabilities
  #   sim        = simulation number.
  
  # Check error dist
  stopifnot(err.dist == "normal" | err.dist == "t")
  
  # set seed
  set.seed(seed)
  
  # Draw errors
  if (err.dist == "normal") {
    sigma.full = matrix(rnorm(Nunits*timePeriods, mean = 0, sd = 1), nrow = Nunits, ncol = timePeriods)
  } else {
    sigma.full = matrix(rt(Nunits*timePeriods, df = 1), nrow = Nunits, ncol = timePeriods)
  }
  
  # Redraw AR1 panel with these simulation errors
  if (parallel) {
    cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
    doParallel::registerDoParallel(cores)
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %dopar% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full, parallel = FALSE)
      tibble(
        plag0 = total.normal.pval(expDF = temp, p.lag = 0)$Z,
        plag1 = total.normal.pval(expDF = temp, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  } else {
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %do% {
      temp = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                               phi = phi, beta = beta, pT = pT, err.dist = err.dist, sigma = sigma, 
                               sigma.full = sigma.full)
      tibble(
        plag0 = total.normal.pval(expDF = temp, p.lag = 0)$Z,
        plag1 = total.normal.pval(expDF = temp, p.lag = 1)$Z,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1],
        phi = phi[1],
        beta = beta[1],
        err.dist = err.dist,
        sim = i
      )
    }
  }
  return(simulated_data)
}

AR1.panel.gen.multiple_largeN_beta0 <- function(sims, timePeriods, Nunits, phi, pT, err.dist, sigma, 
                                                same.err, seed, parallel) {
  # Description
  #   This function computes multiple draws of a panel experiment 
  #   at a given choice of parameters using the same errors. This is
  #   useful for analyzing the properties of the estimator over the randomizaton distribution.
  #
  #   It returns a tibble that contains the time-specific p-lag = 0, 1, 2 estimates in the last period
  # 
  # Args:
  #   sims        = A number indicating the number of paths of the science table to try.
  #   timePeriods = A number indicating the total number of time setps
  #   Nunits      = A number specifying the number of units
  #   phi         = N x 1 number specifying the AR coefficients for each unit
  #   beta        = N x 1 number indicating the unit specific, time invariant contemporaneous treatment effect.
  #   pT          = N x T matrix giving the unit-specific treatment probabilities across periods.
  #   err.dist    = specifies the distribution of errors. Must be either 'normal' or 't'
  #   sigma       = specifies the variance of the gaussian innovations if err.dist = normal or degrees of freedom 
  #                of student t distribution if err.dist = 't'
  #   seed        = seed
  #   parallel    = flag for parallel option with AR1.panel.gen.obs
  #
  # Returns: Returns a sims*(Nunits*timePeriods) x 5 dataframe containing the columns
  #   unitId     = unit level identifiers
  #   Yobs       = observed outcomes
  #   Wobs       = observed treatments
  #   pT         = treatment probabilities
  #   sim        = simulation number.
  
  # Check error dist
  stopifnot(err.dist == "normal" | err.dist == "t")
  
  # set seed
  set.seed(seed)
  
  # Draw errors
  if (err.dist == "normal") {
    sigma.full = matrix(rnorm(Nunits*timePeriods, mean = 0, sd = 1), nrow = Nunits, ncol = timePeriods)
  } else {
    sigma.full = matrix(rt(Nunits*timePeriods, df = 1), nrow = Nunits, ncol = timePeriods)
  }
  
  # Draw initial data
  init_data = AR1.panel.gen.obs(N = Nunits, timePeriods = timePeriods, 
                                phi = phi, beta = rep(0, Nunits), 
                                pT = pT, err.dist = err.dist, 
                                sigma = sigma, 
                                sigma.full = sigma.full, 
                                parallel = FALSE)
  
  # Redraw AR1 panel with these simulation errors
  if (parallel) {
    cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
    doParallel::registerDoParallel(cores)
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %dopar% {
      set.seed(i)
      temp = init_data
      temp$Wobs = rbinom(Nunits*timePeriods, 1, pT[1, 1])
      
      plag0 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 0)$Z
      plag1 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 1)$Z
      
      tibble(
        plag0 = plag0,
        plag1 = plag1,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1, 1],
        phi = phi[1],
        err.dist = err.dist,
        sim = i
      )
    }
  } else {
    simulated_data = foreach(i = 1:sims, .combine = 'rbind') %do% {
      set.seed(i)
      temp = init_data
      temp$Wobs = rbinom(Nunits*timePeriods, 1, pT[1, 1])
      
      plag0 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 0)$Z
      plag1 = time.normal.pval(expDF = temp, time = timePeriods, p.lag = 1)$Z
      
      tibble(
        plag0 = plag0,
        plag1 = plag1,
        timePeriods = timePeriods,
        Nunits = Nunits,
        pT = pT[1, 1],
        phi = phi[1],
        err.dist = err.dist,
        sim = i
      )
    }
  }
  return(simulated_data)
}