# Author: Ashesh Rambachan <asheshr@g.harvard.edu>
# Description:
#   The script is used to simulate results for a given combination of input parameters for the fixed T, large N simulations. The combination is taken from
# panel_param_combinations_largeN.csv according to LSB_JOBINDEX row number. It is then processed and results are outputted 
#  into scrap/sim_res[LSB_JOBINDEX]_largeN.rds
#
# To ensure that this script runs more quickly, this simulation file is structured differently than the others.

# Libraries ------------------
library(foreach)
library(doParallel)

# Helper functions -----------------------
est_plag_qstep_ht <- function(Yobs, Wobs, pT, p.lag, q.step = 0, normalize, pT.treatmentprob) {
  # Description:
  #   This function computes the p-lag q-step estimator of the p-lag q-step 
  #   causal effect. p.lag = 0 corresponds to the contemporaneous causal effect.
  #
  #   This function is copied from the ts_analysis_function.R in Bojinov & Shephard (2019).
  #   It is used throughout the more general functions defined later on. 
  #
  #   Note: We modify this function so that pT only accepts a vector of treatment probabilities.
  #
  #   Note: We mask this function from the user as its outputs can be generated using the function
  #         unit.ht.wavg.est() defined below.
  #
  # Args:
  #   Yobs      = The observed data.
  #   Wobs      = The corresponding treatment assignments.
  #   pT        = A vector containing the probability of receiving treatment at each time poit.
  #   p.lag     = A non-negative integer indicating the p-lag.
  #   q.step    = A positive integer; indicating the number of steps. This is always set to zero.
  #   normalize = A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   pT.treatmentprob = A logical value indicating if pT is the probability of treatment (TRUE)
  #                      or the probability of the observed treatment (FALSE)
  # Returns: List containing elements
  #   est.mean = The mean of the estimated values. 
  #   est      = The estimated value
  
  stopifnot((q.step + p.lag + 1) > 0)
  p.lag <- p.lag + 1
  T <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  if (pT.treatmentprob) {
    pT <- pT * Wobs + (1 - pT) * (1 - Wobs)
  }
  # Compute the probability of the received treatment
  if ((q.step + p.lag) > 1) {
    pT.mat <- matrix(NA, nrow = (q.step + p.lag), ncol = T)
    pT.mat[1, ] <- pT
    for (kk in 1:(q.step + p.lag - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT[1 : (T - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT
  }
  if (p.lag > 0){
    Wobs <- c(rep(0, p.lag - 1), Wobs[1:(T - p.lag + 1)])
  }
  est <- as.numeric(Yobs * (( - 1)^(1 - Wobs)) * ((1 / 2) ^ ((p.lag + q.step) - 1)) / pT.prod)
  if (normalize) {
    gm <- .var_fisher(Yobs, Wobs, pT, p.lag, q.step)$var.est
    # Sometimes with very low probability gm = 0, then you just add some noise 
    # since 0/a = 0, this doesn't change anything 
    gm[which(gm == 0)] <- abs(jitter(gm[which(gm == 0)]))
    est <- est / sqrt(gm)
  }
  est.mean <- mean(est[(p.lag + q.step):T], na.rm = TRUE)
  return(list(est.mean = est.mean, est = est))
}
cons_var_bound <- function(Yobs, Wobs, pT, p.lag, q.step  = 0) { 
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes the conservative variance bound for the p-lag, q-step HT estimator. 
  #   This function is copied from the ts_analysis_function.R in Bojinov & Shephard (2019).
  #
  #   Notes: 
  #     1) We modify this function so that pT only accepts a vector of treatment probabilities.
  #     2) This is given in Eqn (5) of Bojinov Et Al in Section 3.2, which provides the expression for the i,t-th lag-p
  #        HT estimator. It is also given in Lemma 2 in Section B.2 of the online appendix of Bojinov & Shepard (2019).
  #     3) We mask this function from the user as it is used as a helper function
  #        to the functions defined below.
  #
  # Args:
  #   Yobs   = A vector of length T, the observed outcome path.
  #   Wobs   = A vector of treatment assignments
  #   pT     = A vector containing the probability of assigning a unit to treatment.
  #   p.lag  = A non-negative integer indicating the p-lag.
  #   q.step = A positive integer; indicating the number of steps. This is always set to zero.
  # Returns: List containing elements
  #   var.est   =  The estimated variance bound at each time point.
  #   gm.sq.est = The estimate of gm (the average of the variances). 
  
  timePeriods <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  K.step <- p.lag + q.step + 1
  if (K.step > 1) {
    pT.mat <- matrix(NA, nrow = K.step, ncol = timePeriods)
    pT.mat[1, ] <- pT * Wobs + (1 - pT) * (1 - Wobs)
    for (kk in 1:(K.step - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT.mat[1, 1:(timePeriods - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT * Wobs + (1 - pT) * (1 - Wobs)
  }
  var.est <- as.numeric(Yobs^2  * (1 + 2 * pT.prod * (2^(K.step - 1) - 1)) *
                          ((1 / 4)^(K.step - 1)) / pT.prod^2) 
  gm.sq.est <- sum(var.est[K.step:timePeriods]) / (timePeriods - K.step + 1) ^ 2
  return(list(var.est = var.est, gm.sq.est = gm.sq.est))
}

# Simulation parameteres
job_index <- as.integer(Sys.getenv("LSB_JOBINDEX"))
Sys.sleep(runif(1)*10)
combination <- read.csv("data/panel_param_combinations_largeN.csv")[job_index,]

# Draw errors once
set.seed(1)
sigma.full = matrix(rt(combination$Nunits*combination$timePeriods, df = 1),
                    nrow = combination$Nunits, ncol = combination$timePeriods)
Yobs = t(apply(sigma.full, 1, function(x) { as.numeric(stats::filter(x, combination$phi, method = 'recursive')) }))

# Compute variance bound once
Wobs = matrix(rbinom(combination$Nunits*combination$timePeriods, 1, combination$pT), 
              nrow = combination$Nunits, ncol = combination$timePeriods)

varMat_plag0 = t(sapply(1:combination$Nunits, function(x) { cons_var_bound(Yobs[x,], Wobs[x,], combination$pT, p.lag = 0)$var.est }))
sigma_plag0 = sqrt( sum(varMat_plag0[, combination$timePeriods], na.rm = TRUE)/combination$Nunits^2 )

varMat_plag1 = t(sapply(1:combination$Nunits, function(x) { cons_var_bound(Yobs[x,], Wobs[x,], combination$pT, p.lag = 1)$var.est }))
sigma_plag1 = sqrt( sum(varMat_plag1[, combination$timePeriods], na.rm = TRUE)/combination$Nunits^2 )

# Run simulation
cores = parallel::detectCores(all.tests = FALSE, logical = TRUE)
doParallel::registerDoParallel(cores)
resR.temp = foreach(i = 1:combination$numSims, .combine = 'rbind') %dopar% {
  set.seed(i)

  # Draw treatment panel
  Wobs = matrix(rbinom(combination$Nunits*combination$timePeriods, 1, combination$pT), 
                nrow = combination$Nunits, ncol = combination$timePeriods)
  
  # Estimate plag0
  estMat_plag0 = t(sapply(1:combination$Nunits, function(x) { est_plag_qstep_ht(Yobs[x,], Wobs[x,], combination$pT, p.lag = 0, 
                                                                    normalize = FALSE, pT.treatmentprob = TRUE)$est }))
  est_plag0 = sum(estMat_plag0[, combination$timePeriods], na.rm = TRUE)/combination$Nunits
  
  # Estimate plag1
  estMat_plag1 = t(sapply(1:combination$Nunits, function(x) { est_plag_qstep_ht(Yobs[x,], Wobs[x,], combination$pT, p.lag = 1, 
                                                                                normalize = FALSE, pT.treatmentprob = TRUE)$est }))
  est_plag1 = sum(estMat_plag0[, combination$timePeriods], na.rm = TRUE)/combination$Nunits
  
  data.frame(plag0 = est_plag0/sigma_plag0,
             plag1 = est_plag1/sigma_plag1,
             timePeriods = combination$timePeriods,
             Nunits = combination$Nunits,
             pT = combination$pT,
             phi = combination$phi,
             beta = 0,
             err.dist = combination$err.dist,
             sim = i)
}

# Create a temporary directory scrap/ if it doesn't exist
if (!dir.exists("scrap")) {
  dir.create('scrap')
}
path <- paste0("scrap/sim_res", job_index, "_largeN.rds")
saveRDS(resR.temp, path) 

